#include <iostream>
#include <algorithm>
#include <unordered_set>
#include <set>
#include "MeshLoader.h"
#include "Hash.h"
#include <fstream>
#include <ranges>


const std::vector<Node>& MeshLoader::getNodes() {
    return Nodes;
}


const std::vector<FiniteElement>& MeshLoader::getFiniteElements() {
    return FiniteElements;
}


const std::vector<FiniteElement>& MeshLoader::getBoundaryElements() {
    return BoundaryElements;
}


void MeshLoader::printNodes(const std::vector<Node>& p_nodes) {
    std::cout << p_nodes;
}


void MeshLoader::printFiniteElements(const std::vector<FiniteElement>& p_elem) {
    std::cout << p_elem;
}


//    
std::vector<int> MeshLoader::getFiniteElementByID(int p_id1, int p_id2, int p_id3) {
    std::vector<int> res;

    auto findFiniteElement = [&p_id1, &p_id2, &p_id3, &res](const FiniteElement& elem) {

        if ((std::any_of(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(),
            [p_id1](const int id) {return id == p_id1; })) &&     
            (std::any_of(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(),
                [p_id2](const int id) {return id == p_id2; })) &&
            (std::any_of(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(),
                [p_id3](const int id) {return id == p_id3; }))) {
            res.push_back(elem.m_FiniteElementID);
        }
    };

    std::for_each(FiniteElements.begin(), FiniteElements.end(), findFiniteElement);
    return res;
}


//    ()
std::vector<int> MeshLoader::getFiniteElementByEdge(int p_id1, int p_id2) {
    std::vector<int> res;
    auto findFiniteElement = [&p_id1, &p_id2, &res](const FiniteElement& elem) {

        if ((std::any_of(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(),
            [p_id1](const int id) {return id == p_id1; })) && 
            (std::any_of(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(),
                [p_id2](const int id) {return id == p_id2; }))) {
            res.push_back(elem.m_FiniteElementID);
        }
    };

    std::for_each(FiniteElements.begin(), FiniteElements.end(), findFiniteElement);
    return res;
}


//  ID   ID 
std::vector<int> MeshLoader::getNodesIDByBorder(int p_id) { 
    std::set<int> res;
    auto cur = BoundaryElements.begin();

    while (cur != BoundaryElements.end()) {
        cur = std::find_if(cur, BoundaryElements.end(), [p_id](const FiniteElement& elem) {
            return elem.m_GeometryAreaID == p_id;
            }
        );
        if (cur != BoundaryElements.end()) {
            for (const auto& nod : cur->m_NodeIDs) {
                Node nodeId = Nodes[nod - 1];
                res.insert(nodeId.m_ID);
            }
            ++cur;
        }
    }
    return std::vector<int>(res.begin(), res.end());
}


//  ID   ID 
std::vector<int> MeshLoader::getFiniteElementIdBySurface(int p_id) {
    std::vector<int> res;
    auto findFiniteElement = [&p_id, &res](const FiniteElement& elem) {
        if (elem.m_GeometryAreaID == p_id) {
            res.push_back(elem.m_FiniteElementID);
        }
    };

    std::for_each(FiniteElements.begin(), FiniteElements.end(), findFiniteElement);
    return res;
}


//  ID    ID 
std::vector<int> MeshLoader::getBoundElementIDByBorder(int _id) {
    std::vector<int> res;
    auto po = [&_id, &res](const FiniteElement& elem) {
        if (elem.m_GeometryAreaID == _id) {
            res.push_back(elem.m_FiniteElementID);
        }
    };

    std::for_each(BoundaryElements.begin(), BoundaryElements.end(), po);
    return res;
}


//   
std::vector<std::set<int>> MeshLoader::createNeighborsVector() {
    std::vector<std::set<int>> neighbors(Nodes.size()+1);
    for (const auto& elem : BoundaryElements) {
        for (auto nodeId : elem.m_NodeIDs) {
            for (auto anthNodeId : elem.m_NodeIDs)
                if (nodeId != anthNodeId) {
                    neighbors[nodeId].insert(anthNodeId);
                }
        }
    }

    auto it = neighbors.begin();
    std::advance(it, 1);
    int i = 1;
    for (; it != neighbors.end(); ++it) {
        std::cout << "  neighbours of node " << i++ << ": ";
        for (const auto& element : *it) {
            std::cout << element << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;

    return neighbors;
}


//     
std::array<double, 3> MeshLoader::getMiddleCoords(const Edge& edge) {
    double MiddleX = ((Nodes.at(edge.m_edgeNodes.first - 1).m_coordinates.at(0)) +
        (Nodes.at(edge.m_edgeNodes.second - 1).m_coordinates.at(0))) / 2;
    double MiddleY = ((Nodes.at(edge.m_edgeNodes.first - 1).m_coordinates.at(1)) +
        (Nodes.at(edge.m_edgeNodes.second - 1).m_coordinates.at(1))) / 2;
    double MiddleZ = ((Nodes.at(edge.m_edgeNodes.first - 1).m_coordinates.at(2)) +
        (Nodes.at(edge.m_edgeNodes.second - 1).m_coordinates.at(2))) / 2;
    return { MiddleX, MiddleY, MiddleZ };
}


Node MeshLoader::getMiddleNode(const Edge& edge) {
    Node newNode(Nodes.size() + 1, getMiddleCoords(edge), false);
    return newNode;
}


void MeshLoader::insertNodeMid() {
    std::unordered_set<Edge, Hash> edge;
    for (auto& elem : FiniteElements) {
        std::vector<int> ElemNodeID = elem.m_NodeIDs; 
        size_t size = elem.m_NodeIDs.size();
        for (int first = 0; first < size; ++first) { 
            for (auto sec = first + 1; sec < size; ++sec) { 
                Edge curEdge(ElemNodeID[first], ElemNodeID[sec]); 
                Node newNode = getMiddleNode(curEdge); 
                curEdge.m_midNode = newNode.m_ID; 
                if (edge.insert(curEdge).second) { 
                    Nodes.push_back(newNode);
                    elem.m_NodeIDs.push_back(newNode.m_ID); 
                }
                else {
                    elem.m_NodeIDs.push_back(curEdge.m_midNode);
                }
            }
        }
        std::cout << elem << std::endl;
    }
    

    for (auto& belem : BoundaryElements) { 
        std::vector<int> bfeNodesId = belem.m_NodeIDs; 
        size_t size = belem.m_NodeIDs.size(); 
        for (auto first = 0; first < size; ++first) { 
            for (auto sec = first + 1; sec < size; ++sec) { 
                auto curEdgeIter = edge.find({ bfeNodesId[first], bfeNodesId[sec] }); 
                if (curEdgeIter == edge.end()) {  
                    curEdgeIter = edge.find({ bfeNodesId[sec], bfeNodesId[first] });
                }
                belem.m_NodeIDs.push_back(curEdgeIter->m_midNode);
            }
        }
    }
}